from dateutil import rrule
from plone.event.utils import DSTAUTO
from plone.event.utils import dt2int
from plone.event.utils import pydt
from plone.event.utils import tzdel
from plone.event.utils import utc
from plone.event.utils import utcoffset_normalize

import datetime
import re


# TODO: make me configurable
MAXCOUNT = 1000  # Maximum number of occurrences


def recurrence_sequence_ical(
    start,
    recrule=None,
    from_=None,
    until=None,
    count=None,
    duration=None,
):
    """Calculates a sequence of datetime objects from a recurrence rule
    following the RFC2445 specification, using python-dateutil recurrence
    rules.  The resolution of the resulting datetime objects is one second,
    since python-dateutil rrulestr doesn't support microseconds.

    :param start:   datetime or DateTime instance of the date from which the
                    recurrence sequence is calculated.
    :type start: datetime.datetime

    :param recrule: Optional string with RFC2445 compatible recurrence
                    definition, dateutil.rrule or dateutil.rruleset instances.
    :type recrule: string

    :param from_:   Optional datetime or DateTime instance of the date, to
                    limit the result within a timespan.
    :type from_: datetime.datetime

    :param until:   Optional datetime or DateTime instance of the date, until
                    the recurrence is calculated. If not given, count or
                    MAXDATE limit the recurrence calculation.
    :type until: datetime.datetime

    :param count:   Optional integer which defines the number of occurrences.
                    If not given, until or MAXDATE limits the recurrence
                    calculation.
    :type count: integer

    :param duration: Optional timedelta instance, which is used to calculate
                     if a occurrence datetime plus duration is within the
                     queried timerange.
    :type duration:  datetime.timedelta

    :returns: A generator which generates a sequence of datetime instances.
    :rtype: generator

    """
    # Always use python datetime objects and remove the microseconds
    start = pydt(start, exact=False)
    from_ = pydt(from_, exact=False)
    until = pydt(until, exact=False)
    tz = start.tzinfo
    start = tzdel(start)  # tznaive | start defines tz
    _from = tzdel(from_)
    _until = tzdel(until)
    if duration:
        assert isinstance(duration, datetime.timedelta)
    else:
        duration = datetime.timedelta(0)

    if recrule:
        # We want the recurrence be calculated ignoring the DTSTART,
        # which is defined by the event's own start.
        # ‌ Also set the UNTIL time to the end of the day to include the last
        # occurrence for sure.
        #
        # start is a mandatory parameter for this function, remove DTSTART
        # from recrule
        recrule = re.sub(r"DTSTART:[^;\n]*[;\n]", "", recrule, re.MULTILINE)
        # TODO BUGFIX WRONG TIME DEFINITIONS
        # THIS HACK ensures, that UNTIL, RDATE and EXDATE definitions with
        # incorrect time (currently always set to 0:00 by the recurrence
        # widget) are handled correctly.
        #
        # Following fixes are made:
        # - The UNTIL date should be included in the recurrence set, as defined
        #   by RFC5545 (fix sets it to the end of the day)
        # - RDATE definitions should have the same time as the start date.
        # - EXDATE definitions should exclude occurrences on the specific date
        #   only the same time as the start date.
        # In the long term ,the recurrence widget should be able to set the
        # time for UNTIL, RDATE and EXDATE.
        t0 = start.time()  # set initial time information.
        # First, replace all times in the recurring rule with starttime
        t0str = f"T{t0.hour:02d}{t0.minute:02d}{t0.second:02d}"
        # Replace any times set to 000000 with start time, not all
        # rrules are set by a specific broken widget.  Don't waste time
        # subbing if the start time is already 000000.
        if t0str != "T000000":
            recrule = re.sub(r"T000000", t0str, recrule)
        # Then, replace each until times with the end of the day
        recrule = re.sub(
            r"(UNTIL[^T]*[0-9]{8})T([0-9]{6}Z?)",
            r"\1T235959",
            recrule,
        )

        # RFC2445 string
        # forceset: always return a rruleset
        # dtstart: optional used when no dtstart is in RFC2445 string
        #          dtstart is given as timezone naive time. timezones are
        #          applied afterwards, since rrulestr doesn't normalize
        #          timezones over DST boundaries
        rset = rrule.rrulestr(
            recrule,
            dtstart=start,
            forceset=True,
            ignoretz=True,
            # compatible=True # RFC2445 compatibility
        )
    else:
        rset = rrule.rruleset()
    rset.rdate(start)  # RCF2445: always include start date

    # limit
    if _from and _until:
        # between doesn't add a ruleset but returns a list
        rset = rset.between(_from - duration, _until, inc=True)
    for cnt, date in enumerate(rset):
        # Localize tznaive dates from rrulestr sequence
        date = tz.localize(date)

        # Limit number of recurrences otherwise calculations take too long
        if MAXCOUNT and cnt + 1 > MAXCOUNT:
            break
        if count and cnt + 1 > count:
            break
        if from_ and utc(date) + duration < utc(from_):
            continue
        if until and utc(date) > utc(until):
            break

        yield date
    return


def recurrence_sequence_timedelta(
    start,
    delta=None,
    until=None,
    count=None,
    dst=DSTAUTO,
):
    """Calculates a sequence of datetime objects from a timedelta integer,
    which defines the minutes between each occurrence.

    :param start: datetime or DateTime instance of the date from which the
                  recurrence sequence is calculated.
    :type start: datetime

    :param delta: Integer which defines the minutes
                  between each date occurrence.
    :type delta: integer

    :param until: datetime or DateTime instance of the date, until the
                  recurrence is calculated. If not given,
                  count or MAXDATE limit the recurrence calculation.
    :type until: datetime

    :param count: Integer which defines the number of occurrences. If not given,
                  until or MAXDATE limits the recurrence calculation.
    :param count: integer

    :param dst:   Daylight Saving Time crossing behavior. DSTAUTO, DSTADJUST or
                  DSTKEEP. For more information, see
                  plone.event.utils.utcoffset_normalize.
    :param dst: string

    :return: A generator which generates a sequence of datetime instances.
    :rtype: generator

    """
    start = pydt(start)
    yield start

    if delta is None or delta < 1 or until is None:
        return

    until = pydt(until)

    before = start
    delta = datetime.timedelta(minutes=delta)
    cnt = 0
    while True:
        after = before + delta
        after = utcoffset_normalize(after, delta, dst)

        # Limit number of recurrences otherwise calculations take too long
        if MAXCOUNT and cnt + 1 > MAXCOUNT:
            break
        if count and cnt + 1 > count:
            break
        if until and utc(after) > utc(until):
            break
        cnt += 1

        yield after
        before = after


def recurrence_int_sequence(sequence):
    """Generates a sequence of integer representations from a sequence of
    dateime instances.

    :param sequence: An iterable sequence of datetime instances.
    :type sequence: iterable
    :returns: Generator of integer representations of datetime instances.
    :rtype: generator

    """
    for dt in sequence:
        yield dt2int(dt)



import unittest


class TestRecurrenceSequenceIcal(unittest.TestCase):
    def test_start(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        seq = recurrence_sequence_ical(start)
        results = [res for res in seq]
        self.assertEqual(len(results), 1)

    def test_recrule_str(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        recrule = "FREQ=DAILY;INTERVAL=10;COUNT=5"
        seq = recurrence_sequence_ical(start, recrule=recrule)
        results = [res for res in seq]
        self.assertEqual(len(results), 5)

    def test_recrule_str_rdate(self):
        """Test, if an RDATE date has the correct time set.
        See: "BUGFIX WRONG RDATE TIME" in recurrence.py
        """
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23, 10, 10)
        recrule = """FREQ=DAILY;INTERVAL=1;COUNT=3
RDATE:20111129T000000"""
        seq = recurrence_sequence_ical(start, recrule=recrule)
        results = [res for res in seq]
        self.assertEqual(len(results), 4)
        self.assertEqual(results[0].time(), results[-1].time())

    def test_recrule_str_exdate(self):
        """Test, if an EXDATE date are not in the resulting recurrence set."""
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        import pytz

        at = pytz.timezone("Europe/Vienna")
        start = at.localize(datetime(2013, 6, 29, 10, 10))
        recrule = "RRULE:FREQ=DAILY;COUNT=4\r\nEXDATE:20130630T000000,20130701T000000\r\nRDATE:20130706T000000,20130809T000000"  # noqa
        seq = recurrence_sequence_ical(start, recrule=recrule)
        res = [res for res in seq]
        res_test = [
            at.localize(datetime(2013, 6, 29, 10, 10)),
            at.localize(datetime(2013, 7, 2, 10, 10)),
            at.localize(datetime(2013, 7, 6, 10, 10)),
            at.localize(datetime(2013, 8, 9, 10, 10)),
        ]
        self.assertEqual(len(res), 4)
        self.assertEqual(res, res_test)

    def test_recrule_str_until(self):
        """Test, if UNTIL stops the sequence at the end of the day, even if
        it's set to 0:00 by the recurrence widget.
        """
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        import pytz

        at = pytz.timezone("Europe/Vienna")
        start = at.localize(datetime(2013, 6, 29, 10, 10))
        recrule = "RRULE:FREQ=DAILY;UNTIL=20130702T000000"
        seq = recurrence_sequence_ical(start, recrule=recrule)
        res = [res for res in seq]
        res_test = [
            at.localize(datetime(2013, 6, 29, 10, 10)),
            at.localize(datetime(2013, 6, 30, 10, 10)),
            at.localize(datetime(2013, 7, 1, 10, 10)),
            at.localize(datetime(2013, 7, 2, 10, 10)),
        ]
        self.assertEqual(len(res), 4)
        self.assertEqual(res, res_test)

    def test_recrule_from_until(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        recrule = None
        from_ = datetime(2011, 11, 1)
        until = datetime(2011, 12, 31)
        seq = recurrence_sequence_ical(
            start,
            recrule=recrule,
            from_=from_,
            until=until,
        )
        results = [res for res in seq]
        self.assertEqual(len(results), 1)

    def test_recrule_str_more_than_MAXCOUNT(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        recrule = "FREQ=DAILY;INTERVAL=10;COUNT=1001"
        seq = recurrence_sequence_ical(start, recrule=recrule)
        results = [res for res in seq]
        self.assertEqual(len(results), 1000)

    def test_recrule_str_more_than_count(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        recrule = "FREQ=DAILY;INTERVAL=10;COUNT=10"
        count = 5
        seq = recurrence_sequence_ical(start, recrule=recrule, count=count)
        results = [res for res in seq]
        self.assertEqual(len(results), 5)

    def test_recrule_from(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        recrule = "FREQ=DAILY;INTERVAL=1;COUNT=5"
        from_ = datetime(2011, 11, 25)
        until = datetime(2011, 11, 27)
        seq = recurrence_sequence_ical(
            start,
            recrule=recrule,
            from_=from_,
            until=until,
        )
        results = [res for res in seq]
        self.assertEqual(len(results), 3)

    def test_recrule_until(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 24)
        recrule = "FREQ=DAILY;INTERVAL=1;COUNT=5"
        from_ = datetime(2011, 11, 23)
        until = datetime(2011, 11, 27)
        seq = recurrence_sequence_ical(
            start,
            recrule=recrule,
            from_=from_,
            until=until,
        )
        results = [res for res in seq]
        self.assertEqual(len(results), 4)

    def test_recrule_from_until_with_duration(self):
        """Should include events ranging into the queried timerange."""
        from datetime import datetime
        from datetime import timedelta
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 23)
        recrule = "FREQ=DAILY;INTERVAL=1;COUNT=5"
        from_ = datetime(2011, 11, 26)
        until = datetime(2011, 11, 27)
        seq = recurrence_sequence_ical(
            start,
            recrule=recrule,
            from_=from_,
            until=until,
            duration=timedelta(days=2),
        )
        results = [res for res in seq]
        self.assertEqual(len(results), 4)

    def test_recrule_until_with_timezone(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        start = datetime(2011, 11, 24)
        recrule = "RRULE:FREQ=DAILY;UNTIL=20111130T000000Z"
        seq = list(recurrence_sequence_ical(start, recrule=recrule))
        self.assertEqual(len(seq), 7)

    def test_recrule_with_dtstart(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_ical

        import pytz

        at = pytz.timezone("Europe/Vienna")
        start = at.localize(datetime(2023, 9, 4, 1, 0))
        # DTSTART is ignored, because start is ever explicitly given
        recrule = "DTSTART:20230903T180000Z\nRRULE:FREQ=DAILY;UNTIL=20230905T230000Z"
        seq = list(recurrence_sequence_ical(start, recrule=recrule))
        self.assertEqual(len(seq), 2)
        self.assertEqual(seq[0], at.localize(datetime(2023, 9, 4, 1, 0)))
        self.assertEqual(seq[1], at.localize(datetime(2023, 9, 5, 1, 0)))



from plone.event.interfaces import IEvent
from plone.event.interfaces import IEventAccessor
from zope.component import adapter
from zope.interface import implementer


@implementer(IEventAccessor)
@adapter(IEvent)
class EventAccessor:
    """Simple event accessor adapter implementation for generic events, which
    follow the IEvent interface closely.

    Concrete implementations adapt a content type to the IEvent specification.
    """

    def __init__(self, context):
        object.__setattr__(self, "context", context)

    def __getattr__(self, name):
        return getattr(self.context, name)

    def __setattr__(self, name, value):
        setattr(self.context, name, value)

    def __delattr__(self, name):
        delattr(self.context, name)

    @property
    def duration(self):
        return self.end - self.start



from zope.component.testing import tearDown

import doctest
import os.path
import unittest


OPTIONFLAGS = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
DOCFILES = [
    "recurrence.rst",
    "recurrence_dateutil.rst",
    "utils.rst",
]
DOCMODS = [
    "plone.event.utils",
]


def test_suite():
    suite = unittest.TestSuite()
    suite.addTests(
        [
            doctest.DocFileSuite(
                os.path.join(os.path.dirname(__file__), "..", docfile),
                module_relative=False,
                optionflags=OPTIONFLAGS,
                tearDown=tearDown,
            )
            for docfile in DOCFILES
        ]
    )
    suite.addTests(
        [doctest.DocTestSuite(docmod, optionflags=OPTIONFLAGS) for docmod in DOCMODS]
    )
    return suite


if __name__ == "__main__":
    unittest.main(defaultTest="test_suite")



from unittest import mock

import unittest


class TestUtils(unittest.TestCase):
    @mock.patch("plone.event.utils.pytz")
    @mock.patch("plone.event.utils.os")
    def test_default_timezone(self, os, pytz):
        from plone.event.utils import default_timezone

        os.environ.keys.return_value = ["TZ"]
        os.environ = mock.MagicMock()
        pytz.timezone().zone = "zone"
        self.assertEqual(default_timezone(), "zone")

    def test_utcoffset_normalize(self):
        from plone.event.utils import utcoffset_normalize

        date = mock.Mock()
        date.replace = mock.Mock(side_effect=KeyError)
        self.assertEqual(utcoffset_normalize(date), date)

    @mock.patch("plone.event.utils.guesstz")
    @mock.patch("plone.event.utils.utctz")
    def test_pydt__missing_zone_is_None(self, utctz, guesstz):
        from plone.event.utils import pydt

        dt = mock.Mock()
        dt.toZone.return_value = dt
        dt.parts.return_value = (2011, 11, 24, 11, 39, 00)
        guesstz.return_value = None
        import pytz

        utctz.return_value = pytz.timezone("UTC")
        pydt(dt)
        self.assertTrue(utctz.called)

    @mock.patch("plone.event.utils.guesstz")
    @mock.patch("plone.event.utils.utctz")
    def test_pydt__missing_zone_is_not_None(self, utctz, guesstz):
        from plone.event.utils import pydt

        dt = mock.Mock()
        import pytz

        utctz.return_value = pytz.timezone("UTC")
        missing_zone = utctz()
        dt.toZone.return_value = dt
        dt.parts.return_value = (2011, 11, 24, 11, 39, 00)
        guesstz.return_value = None
        pydt(dt, missing_zone=missing_zone)
        self.assertEqual(utctz.call_count, 1)

    def test_pydt__wrong_type(self):
        from plone.event.utils import pydt

        self.assertEqual(pydt("wrongtype"), None)

    def test_dt2int_dt_is_None(self):
        from plone.event.utils import dt2int

        self.assertFalse(dt2int(None))

    @mock.patch("plone.event.utils.MAX32", 0)
    @mock.patch("plone.event.utils.utc")
    def test_dt2int_less_MAX32(self, utc):
        from plone.event.utils import dt2int

        dt = mock.Mock()
        dd = mock.Mock()
        utc.return_value = dd
        dd.year = 2011
        dd.month = 11
        dd.day = 24
        dd.hour = 14
        dd.minute = 16
        self.assertRaises(OverflowError, lambda: dt2int(dt))

    @mock.patch("plone.event.utils.MAX32", 1077778937)
    @mock.patch("plone.event.utils.utc")
    def test_dt2int_more_MAX32(self, utc):
        from plone.event.utils import dt2int

        dt = mock.Mock()
        dd = mock.Mock()
        utc.return_value = dd
        dd.year = 2011
        dd.month = 11
        dd.day = 24
        dd.hour = 14
        dd.minute = 16
        value = 1077778936
        self.assertEqual(dt2int(dt), value)



import unittest


class TestRecurrenceSequenceTimedelta(unittest.TestCase):
    def test_delta_None(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_timedelta

        start = datetime(2011, 11, 23)
        until = datetime(2011, 11, 24)
        td = recurrence_sequence_timedelta(start, until=until)
        results = [res for res in td]
        self.assertEqual(len(results), 1)

    def test_delta_zero(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_timedelta

        start = datetime(2011, 11, 23)
        delta = 0
        until = datetime(2011, 11, 24)
        td = recurrence_sequence_timedelta(start, delta=delta, until=until)
        results = [res for res in td]
        self.assertEqual(len(results), 1)

    def test_until_None(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_timedelta

        start = datetime(2011, 11, 23)
        delta = 1
        td = recurrence_sequence_timedelta(start, delta=delta)
        results = [res for res in td]
        self.assertEqual(len(results), 1)

    def test_delta_an_hour_until_next_day(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_timedelta

        start = datetime(2011, 11, 23)
        delta = 60
        until = datetime(2011, 11, 24)
        td = recurrence_sequence_timedelta(start, delta=delta, until=until)
        results = [res for res in td]
        self.assertEqual(len(results), 25)

    def test_recur_more_than_MAXCOUNT(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_timedelta

        start = datetime(2011, 11, 23)
        delta = 1
        until = datetime(2012, 11, 23)
        td = recurrence_sequence_timedelta(start, delta=delta, until=until)
        results = [res for res in td]
        self.assertEqual(len(results), 1001)

    def test_recur_more_than_count(self):
        from datetime import datetime
        from plone.event.recurrence import recurrence_sequence_timedelta

        start = datetime(2011, 11, 23)
        delta = 1
        until = datetime(2011, 11, 24)
        count = 20
        td = recurrence_sequence_timedelta(
            start,
            delta=delta,
            until=until,
            count=count,
        )
        results = [res for res in td]
        self.assertEqual(len(results), 21)



from unittest import mock

import unittest


class TestRecurrenceIntSequence(unittest.TestCase):
    @mock.patch("plone.event.recurrence.dt2int")
    def test_recrule_str_(self, dt2int):
        from plone.event.recurrence import recurrence_int_sequence

        sequence = [1, 2, 3]
        dt2int.return_value = "a"
        res = [a for a in recurrence_int_sequence(sequence)]
        self.assertEqual(res, ["a", "a", "a"])



from zope.interface import Attribute
from zope.interface import Interface


class IEvent(Interface):
    """Event type marker interface."""


class IEventRecurrence(Interface):
    """Event type marker interface for events with an RFC5545 compatible
    recurrence definition.

    """


class IOccurrence(Interface):
    """Marker interface for an occurrence item, which represents a single
    occurrence.

    """

    start = Attribute("Occurrence start date")
    end = Attribute("Occurrence end date")


class ICalendarAccessor(Interface):
    """Generic calendar accessor adapter interface.

    A calendar is a collection of calendar components, like events.

    """

    uid = Attribute("UID of the calendar. Autogenerated, read only.")
    # rw attributes
    title = Attribute("Calendar title.")
    description = Attribute("Calendar description text.")
    timezone = Attribute("Default timezone of the calendar.")

    def items(self):
        """Return a list of calendar component items, like events."""


class IEventAccessor(Interface):
    """Event type generic accessor adapter interface.

    A specific adapter implementation should be provided by the individual
    content type packages.

    """

    # ro attributes
    uid = Attribute("UID of the event. Autogenerated, read only.")
    created = Attribute("Python datetime of the event creation timestamp.")
    duration = Attribute("Duration of the event. Computed attribute.")

    # reference
    url = Attribute(
        "Canonical, unique url of the event. External events "
        "are referenced by the origin url unless explicitly set."
    )

    # rw attributes
    last_modified = Attribute("Last modified Python datetime.")
    title = Attribute("Event title.")
    description = Attribute("Event description text.")
    start = Attribute("Event start date as Python datetime.")
    end = Attribute("Event end date as Python datetime.")
    whole_day = Attribute("Event lasts whole day.")
    open_end = Attribute("Event has no defined end time.")
    timezone = Attribute("Timezone of the event. A pytz timezone identifier.")
    recurrence = Attribute("RFC5545 compatible recurrence definition.")
    location = Attribute("Location of the event.")
    attendees = Attribute("List of attendees.")
    contact_name = Attribute("Contact name.")
    contact_email = Attribute("Contact email.")
    contact_phone = Attribute("Contact phone.")
    event_url = Attribute("Website of the event.")
    subjects = Attribute("Categories.")
    text = Attribute("Body text of the event.")


class IRecurrenceSupport(Interface):
    """Event type recurrence adapter."""

    def occurrences(self, range_start, range_end):
        """Return a list of IOccurrence objects with custom attributes of the
        specific occurrence set.

        :param range_start: Search for occurrences after this date.
        :type range_start: Python datetime.

        :param range_end: Search for occurrences before this date.
        :type range_end: Python datetime.

        """


class IICalendar(Interface):
    """Adapter, which is used to construct an icalendar object."""


class IICalendarEventComponent(Interface):
    """Adapter, which is used to construct an event component object for
    icalendar.

    """


class IVEvent(Interface):
    """RFC5545 Event schema

    ; The following are REQUIRED,
    ; but MUST NOT occur more than once.
    ;
    dtstamp / uid /
    ;
    ; The following is REQUIRED if the component
    ; appears in an iCalendar object that doesn't
    ; specify the "METHOD" property; otherwise, it
    ; is OPTIONAL; in any case, it MUST NOT occur
    ; more than once.
    ;
    dtstart /
    ;
    ; The following are OPTIONAL,
    ; but MUST NOT occur more than once.
    ;
    class / created / description / geo /
    last-mod / location / organizer / priority /
    seq / status / summary / transp /
    url / recurid /
    ;
    ; The following is OPTIONAL,
    ; but SHOULD NOT occur more than once.
    ;
    rrule /
    ;
    ; Either 'dtend' or 'duration' MAY appear in
    ; a 'eventprop', but 'dtend' and 'duration'
    ; MUST NOT occur in the same 'eventprop'.
    ;
    dtend / duration /
    ;
    ; The following are OPTIONAL,
    ; and MAY occur more than once.
    ;
    attach / attendee / categories / comment /
    contact / exdate / rstatus / related /
    resources / rdate / x-prop / iana-prop

    """

    dtstart = Attribute("Start Date/Time")
    dtend = Attribute("End Date/Time")
    duration = Attribute("Duration")
    rrule = Attribute("Recurrence Rule")
    description = Attribute("Description")
    location = Attribute("Location")
    summary = Attribute("Summary")
    url = Attribute("Url")
    attendee = Attribute("Attendee")
    categories = Attribute("Categories")
    contact = Attribute("Contact")

    exdate = Attribute("Exdate")
    rdate = Attribute("Rdate")

    dtstamp = Attribute("Timestamp")
    uid = Attribute("Unique identifier")
    klass = Attribute("Class")  # class
    created = Attribute("Created")
    geo = Attribute("Geo")
    last_mod = Attribute("Last Modified")  # last-mod
    organizer = Attribute("Organizer")
    priority = Attribute("Priority")
    seq = Attribute("Seq")
    status = Attribute("Status")
    transp = Attribute("Transp")
    recurid = Attribute("Recurid")
    attach = Attribute("Attach")
    comment = Attribute("Comment")
    rstatus = Attribute("Rstatus")
    related = Attribute("Related")
    resources = Attribute("Resources")
    x_prop = Attribute("X Prop")  # x-prop
    iana_prop = Attribute("Iana Prop")  # iana-prop



from datetime import date
from datetime import datetime
from datetime import timedelta

import logging
import os
import pytz
import time


DSTADJUST = "adjust"
DSTKEEP = "keep"
DSTAUTO = "auto"
MAX32 = int(2**31 - 1)

logger = logging.getLogger("plone.event")


def validated_timezone(timezone, fallback=None):
    """Validate a given timezone identifier. If a fallback is given, return it
        when the given timezone is not a valid pytz zone. Else raise an
        ValueError exception.

    :param timezone: Timezone identifier to be validated against pytz.
    :type timezone: string

    :param fallback: A fallback timezone identifier.
    :type fallback: string

    :returns: A valid pytz timezone identifier.
    :rtype: string
    :raises: ValueError

    >>> from plone.event.utils import validated_timezone

    Validate a valid timezone:
    >>> validated_timezone('Europe/Vienna')
    'Europe/Vienna'

    Validate an invalid timezone with fallback:
    >>> validated_timezone('NOTVALID', 'UTC')
    'UTC'

    Validate an invalid timezone without fallback:
    >>> validated_timezone('NOTVALID')
    Traceback (most recent call last):
    ...
    ValueError: The timezone NOTVALID ...

    The fallback itself isn't validated:
    >>> validated_timezone('NOTVALID', 'NOTVALID')
    'NOTVALID'

    """
    try:
        # following statement ensures, that timezone is a valid pytz/Olson zone
        return pytz.timezone(timezone).zone
    except Exception:
        if fallback:
            logger.warning(
                "The timezone {} is not a valid timezone from the "
                "Olson database or pytz. Falling back to {}.".format(
                    timezone,
                    fallback,
                )
            )
            return fallback
        else:
            raise ValueError(
                "The timezone {} is not a valid timezone from "
                "the Olson database or pytz.".format(timezone)
            )


def default_timezone(fallback="UTC"):
    """Retrieve the timezone from the server.
    Default Fallback: UTC

    :param fallback: A fallback timezone identifier.
    :type fallback: string

    :returns: A timezone identifier.
    :rtype: string

    >>> from plone.event.utils import default_timezone
    >>> import os
    >>> import time
    >>> timetz = time.tzname
    >>> ostz = 'TZ' in os.environ.keys() and os.environ['TZ'] or None

    >>> os.environ['TZ'] = "Europe/Vienna"
    >>> default_timezone()
    'Europe/Vienna'

    Timezone from time module
    >>> os.environ['TZ'] = ""
    >>> time.tzname = ('CET', 'CEST')
    >>> default_timezone()
    'CET'

    Invalid timezone
    >>> os.environ['TZ'] = "PST"
    >>> default_timezone()
    'UTC'

    Invalid timezone with defined fallback
    >>> os.environ['TZ'] = ""
    >>> time.tzname = None
    >>> default_timezone(fallback='CET')
    'CET'

    Restore the system timezone
    >>> time.tzname = timetz
    >>> if ostz:
    ...     os.environ['TZ'] = ostz
    ... else:
    ...     del os.environ['TZ']

    """

    timezone = None
    if "TZ" in os.environ.keys():
        # Timezone from OS env var
        timezone = os.environ["TZ"]
    if not timezone:
        # Timezone from python time
        zones = time.tzname
        if zones and len(zones) > 0:
            timezone = zones[0]
        else:
            # Default fallback = UTC
            logger.warning(
                "Operating system's timezone cannot be found. Falling back to UTC."
            )
    return validated_timezone(timezone, fallback)


# Display helpers
def is_same_time(start, end, exact=False):
    """Test if event starts and ends at same time.

    :param start: The start datetime.
    :type start: Python datetime or Zope DateTime
    :param end: The end datetime.
    :type end: Python datetime or Zope DateTime
    :param exact: If True, the resolution goes down to microseconds. If False,
                  the resolution are seconds. Default is False.
    :type exact: Boolean
    :returns: True, if start and end have the same time, otherwise False.
    :rtype: Boolean.

    >>> from plone.event.utils import is_same_time
    >>> from datetime import datetime, timedelta

    >>> is_same_time(datetime.now(), datetime.now()+timedelta(hours=1))
    False

    >>> is_same_time(datetime.now(), datetime.now()+timedelta(days=1))
    True

    Resolution is one second
    >>> is_same_time(datetime(2013, 5, 21, 10, 59, 58),
    ...              datetime(2013, 5, 21, 10, 59, 59),
    ...              exact=False)
    False

    Exact:
    >>> now = datetime.now()
    >>> is_same_time(now, now, exact=True)
    True

    """
    start = pydt(start, exact=exact).time()
    end = pydt(end, exact=exact).time()
    if exact:
        return start == end
    else:
        return (
            start.hour == end.hour
            and start.minute == end.minute
            and start.second == end.second
        )


def is_same_day(start, end):
    """Test if event starts and ends at same day.

    >>> from plone.event.utils import is_same_day, utc
    >>> from datetime import datetime, timedelta

    >>> is_same_day(
    ...     datetime(2013, 11, 6, 10, 0, 0),
    ...     datetime(2013, 11, 6, 10, 0, 0) + timedelta(hours=1)
    ... )
    True

    >>> is_same_day(
    ...     datetime(2013, 11, 6, 10, 0, 0),
    ...     datetime(2013, 11, 6, 10, 0, 0) + timedelta(days=1)
    ... )
    False

    >>> is_same_day(datetime(2011, 11, 11, 0, 0, 0,),
    ...             datetime(2011, 11, 11, 23, 59, 59))
    True

    Now with one localized (UTC) datetime:
    >>> is_same_day(
    ...     utc(datetime(2013, 11, 6, 10, 0, 0)),
    ...     datetime(2013, 11, 6, 10, 0, 0)
    ... )
    True
    """
    start = pydt(start)
    end = pydt(end)
    return start.date() == end.date()


# Timezone helpers
def utctz():
    """Return the UTVC zone as a pytz.UTC instance.

    >>> from plone.event.utils import utctz
    >>> utctz()
    <UTC>

    """
    return pytz.timezone("UTC")


def utc(dt):
    """Convert Python datetime to UTC.

    >>> from datetime import datetime
    >>> from plone.event.utils import utc
    >>> utc(datetime(2011,11,11,11,11))
    datetime.datetime(2011, 11, 11, 11, 11, tzinfo=<UTC>)

    >>> import pytz
    >>> at = pytz.timezone('Europe/Vienna')
    >>> dta = datetime(2011,11,11,11,11,tzinfo=at)
    >>> utc(dta)
    datetime.datetime(2011, 11, 11, 10, 11, tzinfo=<UTC>)

    utc'ing None returns None
    >>> utc(None)==None
    True

    """
    if dt is None:
        return None
    dt = pydt(dt)
    if dt is None:
        return None
    return dt.astimezone(utctz())


def utcoffset_normalize(date, delta=None, dstmode=DSTAUTO):
    """Fixes invalid UTC offsets from recurrence calculations.

    :param date: datetime instance to normalize.

    :param delta: datetime.timedelta instance.
                  Mode DSTADJUST: When crossing daylight saving time changes,
                  the start time of the date before DST change will be the same
                  in value as afterwards.  It is adjusted relative to UTC. So
                  8:00 GMT+1 before will also result in 8:00 GMT+2 afterwards.
                  This is what humans might expect when recurring rules are
                  defined.
                  Mode DSTKEEP: When crossing daylight saving time changes, the
                  start time of the date before and after DST change will be
                  the same relative to UTC.  So, 8:00 GMT+1 before will result
                  in 7:00 GMT+2 afterwards. This behavior might be what
                  machines expect, when recurrence rules are defined.
                  Mode DSTAUTO: If the relative delta between two occurrences of
                  a reucurrence sequence is less than a day, DSTKEEP will be
                  used - otherwise DSTADJUST. This behavior is the default.
    """
    try:
        assert bool(date.tzinfo)
    except Exception:
        raise TypeError("Cannot normalize timezone naive dates")
    assert dstmode in [DSTADJUST, DSTKEEP, DSTAUTO]
    if delta:
        assert isinstance(delta, timedelta)  # Easier in Java
        delta = delta.seconds + delta.days * 24 * 3600  # total delta in secs
        if dstmode == DSTAUTO and delta < 24 * 60 * 60:
            dstmode = DSTKEEP
        elif dstmode == DSTAUTO:
            dstmode = DSTADJUST

    try:
        if dstmode == DSTKEEP:
            return date.tzinfo.normalize(date)
        else:  # DSTADJUST
            return date.replace(tzinfo=date.tzinfo.normalize(date).tzinfo)
    except Exception:
        # TODO: python-datetime converts e.g RDATE:20100119T230000Z to
        # datetime.datetime(2010, 1, 19, 23, 0, tzinfo=tzutc())
        # should that be a real utc zoneinfo?
        # anyways, return UTC date as-is
        return date


def tzdel(dt):
    """Create timezone naive datetime from a timezone aware one by removing
    the timezone component.

    >>> from plone.event.utils import tzdel, utctz
    >>> from datetime import datetime
    >>> dt = utctz().localize(datetime(2011, 5, 21, 12, 25))

    Remove the timezone:
    >>> tzdel(dt)
    datetime.datetime(2011, 5, 21, 12, 25)

    Using tzdel on a dt instance doesn't alter it:
    >>> dt
    datetime.datetime(2011, 5, 21, 12, 25, tzinfo=<UTC>)

    """
    if dt:
        return dt.replace(tzinfo=None)
    else:
        return None


def is_date(value):
    """Checks, if given value is a date.

    :param value: The value to check.
    :type value: object
    :returns: True, if value is a date (and not a datetime), false otherwise.
    :rtype: Boolean

    >>> from plone.event.utils import is_date
    >>> from datetime import datetime, date
    >>> is_date(date.today())
    True
    >>> is_date(datetime.now())
    False
    >>> is_date(42)
    False
    """
    return type(value) is date


def is_datetime(value):
    """Checks, if given value is a datetime.

    :param value: The value to check.
    :type value: object
    :returns: True, if value is a datetime (and not a date), false otherwise.
    :rtype: Boolean

    >>> from plone.event.utils import is_datetime
    >>> from datetime import datetime, date
    >>> is_datetime(date.today())
    False
    >>> is_datetime(datetime.now())
    True
    >>> is_datetime(42)
    False
    """
    return type(value) is datetime


def date_to_datetime(value):
    """Converts date objects to datetime objects.

    :param value: Date to convert to datetime.
    :type value: date
    :returns: datetime.
    :rtype: datetime

    >>> from plone.event.utils import date_to_datetime
    >>> from datetime import datetime, date
    >>> date_to_datetime(date(2013,3,25))
    datetime.datetime(2013, 3, 25, 0, 0)

    >>> date_to_datetime(datetime(2013,3,25,10,10,10))
    datetime.datetime(2013, 3, 25, 10, 10, 10)

    >>> date_to_datetime(42)
    Traceback (most recent call last):
    ...
    ValueError: Value must be a date or datetime object.


    """

    if is_date(value):
        return datetime(value.year, value.month, value.day)
    elif is_datetime(value):
        return value
    else:
        raise ValueError("Value must be a date or datetime object.")


def pydt(dt, missing_zone=None, exact=False):
    """Converts a Zope's Products.DateTime in a Python datetime.

    :param dt: date, datetime or DateTime object
    :type dt: Python date, datetime or Zope DateTime
    :param missing_zone: A pytz zone to be used, if no timezone is present.
    :type missing_zone: String
    :param exact: If True, the resolution goes down to microseconds. If False,
                  the resolution are seconds. Default is False.
    :type exact: Boolean
    :returns: Python datetime with timezone information.
    :rtype: Python datetime

    >>> from plone.event.utils import pydt
    >>> from datetime import date, datetime
    >>> import pytz

    >>> at = pytz.timezone('Europe/Vienna')
    >>> pydt(at.localize(datetime(2010,10,30)))
    datetime.datetime(2010, 10, 30, 0, 0, tzinfo=<DstTzInfo 'Europe/Vienna' CEST+2:00:00 DST>)

    >>> pydt(date(2010,10,30))
    datetime.datetime(2010, 10, 30, 0, 0, tzinfo=<UTC>)

    pytz cannot handle GMT offsets.
    >>> from DateTime import DateTime
    >>> pydt(DateTime('2011/11/11 11:11:11 GMT+1'))
    datetime.datetime(2011, 11, 11, 10, 11, 11, tzinfo=<UTC>)

    >>> pydt(DateTime('2011/11/11 11:11:11 Europe/Vienna'))
    datetime.datetime(2011, 11, 11, 11, 11, 11, tzinfo=<DstTzInfo 'Europe/Vienna' CET+1:00:00 STD>)

    >>> pydt(DateTime('2005/11/07 18:00:00 Brazil/East'))
    datetime.datetime(2005, 11, 7, 18, 0, tzinfo=<DstTzInfo 'Brazil/East' ...-1 day, 22:00:00 DST>)

    Test with exact set to True
    >>> pydt(DateTime('2012/10/10 10:10:10.123456 Europe/Vienna'), exact=True)
    datetime.datetime(2012, 10, 10, 10, 10, 10, 123456, tzinfo=<DstTzInfo 'Europe/Vienna' CEST+2:00:00 DST>)

    Test with exact set to False
    >>> pydt(DateTime('2012/10/10 10:10:10.123456 Europe/Vienna'), exact=False)
    datetime.datetime(2012, 10, 10, 10, 10, 10, tzinfo=<DstTzInfo 'Europe/Vienna' CEST+2:00:00 DST>)

    >>> pydt(datetime(2012, 10, 10, 20, 20, 20, 123456, tzinfo=at), exact=False)
    datetime.datetime(2012, 10, 10, 20, 20, 20, tzinfo=<DstTzInfo 'Europe/Vienna' CEST+2:00:00 DST>)

    """
    if dt is None:
        return None

    ret = None

    if missing_zone is None:
        missing_zone = utctz()

    if isinstance(dt, date) and not isinstance(dt, datetime):
        dt = datetime(dt.year, dt.month, dt.day)

    if isinstance(dt, datetime):
        tznaive = not bool(getattr(dt, "tzinfo", False))
        if tznaive:
            ret = missing_zone.localize(dt)
        else:
            ret = utcoffset_normalize(dt, dstmode=DSTADJUST)

    if "DateTime" in str(dt.__class__):
        # Zope DateTime
        # TODO: do we need to support subclasses of DateTime too? the check
        #       above would fail.
        tz = guesstz(dt)
        if tz is None:
            dt = dt.toZone(missing_zone.zone)
            tz = missing_zone

        year, month, day, hour, min, sec = dt.parts()[:6]

        # seconds (parts[6]) is a float, so we do modulo for microseconds and
        # map then to int
        #
        # there are precision problems when doing the modulo math
        # (Python 2.7.3 on Ubuntu 12.04):
        # >>> 10.123456%1*1000000
        # 123455.99999999913
        # >>> round(10.123456%1*1000000,0)
        # 123456.0
        micro = int(round(sec % 1 * 1000000))
        sec = int(sec)

        # There is a problem with timezone Europe/Paris
        # tz is equal to <DstTzInfo 'Europe/Paris' PMT+0:09:00 STD>
        dt = datetime(year, month, day, hour, min, sec, micro, tzinfo=tz)
        # before:
        # datetime.datetime(2011, 3, 14, 14, 19,
        # tzinfo=<DstTzInfo 'Europe/Paris' PMT+0:09:00 STD>)
        # dt = dt.tzinfo.normalize(dt)
        # after: datetime.datetime(2011, 3, 14, 15, 10,
        # tzinfo=<DstTzInfo 'Europe/Paris' CET+1:00:00 STD>
        dt = utcoffset_normalize(dt, dstmode=DSTADJUST)
        # after: datetime.datetime(2011, 3, 14, 19,
        # tzinfo=<DstTzInfo 'Europe/Paris' CET+1:00:00 STD>
        ret = dt

    if ret and exact is False:
        ret = ret.replace(microsecond=0)

    return ret


def guesstz(DT):
    """'Guess' pytz from a zope DateTime.

    !!! there is no real good method to guess the timezone.
    DateTime was build somewhere in 1998 long before python had a working
    datetime implementation available and still stucks with this incomplete
    implementation.

    >>> from DateTime import DateTime
    >>> from plone.event.utils import guesstz

    Timezones with the same name as in the Olson DB can easily be guessed.
    >>> guesstz(DateTime('2010-01-01 Europe/Vienna')).zone
    'Europe/Vienna'

    GMT timezones which are popular with DateTime cannot be guessed,
    unfortunately
    >>> guesstz(DateTime('2010-01-01 GMT+1'))
    """
    tzname = DT.timezone()

    # Please note, the GMT offset based timezone information in DateTime are
    # not compatible with Etc/GMT based from pytz. They have different offsets.
    try:
        tz = pytz.timezone(tzname)
        return tz
    except KeyError:
        pass
    return None


# Date as integer representation helpers
def dt2int(dt):
    """Calculates an integer from a datetime, resolution is one minute.
    The datetime is always converted to the UTC zone.

    >>> from plone.event import utils
    >>> from datetime import datetime
    >>> utils.dt2int(datetime(2011,11,11,11,11,tzinfo=utils.utctz()))
    1077760031

    """
    if dt is None:
        return 0
    # TODO: if dt has not timezone information, guess and set it
    dt = utc(dt)
    value = (((dt.year * 12 + dt.month) * 31 + dt.day) * 24 + dt.hour) * 60 + dt.minute

    # TODO: unit test me
    if value > MAX32:
        # value must be integer fitting in the 32bit range
        raise OverflowError(
            """{} is not within the range of indexable dates,<<
            exceeding 32bit range.""".format(
                dt
            )
        )
    return value


def int2dt(dtint):
    """Returns a datetime object from an integer representation with
    resolution of one minute. The datetime returned is in the UTC zone.

    >>> from plone.event.utils import int2dt
    >>> int2dt(1077760031)
    datetime.datetime(2011, 11, 11, 11, 11, tzinfo=<UTC>)

    Dateconversion with int2dt from anything else than integers does not work
    >>> int2dt(.0)
    Traceback (most recent call last):
    ...
    ValueError: int2dt expects integer values as arguments.

    """
    if not isinstance(dtint, int):
        raise ValueError("int2dt expects integer values as arguments.")
    minutes = dtint % 60
    hours = dtint // 60 % 24
    days = dtint // 60 // 24 % 31
    months = dtint // 60 // 24 // 31 % 12
    years = dtint // 60 // 24 // 31 // 12
    return datetime(years, months, days, hours, minutes, tzinfo=utctz())


def dt_to_zone(dt, tzstring):
    """Return a datetime instance converted to the timezone given by the
    string.

    """
    return dt.astimezone(pytz.timezone(tzstring))


# RFC2445 export helpers
def rfc2445dt(dt, mode="utc", date=True, time=True):
    """Convert a datetime or DateTime object into an RFC2445 compatible
    datetime string.

    @param dt: datetime or DateTime object to convert.

    @param mode: Conversion mode ('utc'|'local'|'float')
        Mode 'utc':   Return datetime string in UTC
        Mode 'local': Return datetime string as local
                      including a TZID component
        Mode 'float': Return datetime string as floating (local without TZID
                      component)

    @param date: Return date.

    @param time: Return time.

    Usage
    =====

    >>> from datetime import datetime
    >>> import pytz # this import actually takes quite a long time!
    >>> from plone.event.utils import rfc2445dt

    >>> at = pytz.timezone('Europe/Vienna')
    >>> dt = at.localize(datetime(2010,10,10,10,10))
    >>> dt
    datetime.datetime(2010, 10, 10, 10, 10, tzinfo=<DstTzInfo 'Europe/Vienna' CEST+2:00:00 DST>)

    >>> assert(rfc2445dt(dt) == rfc2445dt(dt, mode='utc'))
    >>> rfc2445dt(dt)
    '20101010T081000Z'

    >>> rfc2445dt(dt, mode='local')
    ('20101010T101000', 'Europe/Vienna')

    >>> rfc2445dt(dt, mode='float')
    '20101010T101000'

    >>> assert(rfc2445dt(dt, date=True, time=True) == rfc2445dt(dt))
    >>> rfc2445dt(dt, time=False)
    '20101010Z'
    >>> rfc2445dt(dt, date=False)
    '081000Z'

    RFC2445 dates from DateTime objects
    -----------------------------------
    >>> from DateTime import DateTime

    It's summer time! So TZ in Belgrade is GMT+2.
    >>> rfc2445dt(DateTime('2010/08/31 18:00:00 Europe/Belgrade'))
    '20100831T160000Z'

    GMT offsets are converted to UTC without any DST adjustments.
    >>> rfc2445dt(DateTime('2010/08/31 20:15:00 GMT+1'))
    '20100831T191500Z'

    """
    # TODO: rfc2445dt might not be necessary. drop me then.

    dt = pydt(dt)
    if mode == "utc":
        dt = utc(dt)
    date = "{}{}{}{}".format(
        date and dt.strftime("%Y%m%d") or "",
        date and time and "T" or "",
        time and dt.strftime("%H%M%S") or "",
        mode == "utc" and "Z" or "",
    )
    if mode == "local":
        return date, dt.tzinfo.zone
    return date



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "2.0.1"

long_description = (
    f"{Path('README.rst').read_text()}\n{Path('CHANGES.rst').read_text()}"
)

setup(
    name="plone.event",
    version=version,
    description="Event and calendaring related tools not bound to Plone",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="Plone calendar calendaring event recurring",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://github.com/plone/plone.event",
    license="GPL",
    packages=find_packages(),
    namespace_packages=["plone"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "setuptools",
        "python-dateutil",  # >4.0.2
        "pytz",
        "zope.component",
        "zope.interface",
    ],
    extras_require={
        "test": [
            "DateTime",
            "zope.configuration",
        ],
    },
)



from datetime import datetime
from plone.event.interfaces import IEvent
from plone.event.interfaces import IEventAccessor
from zope.configuration import xmlconfig

import pytz
import unittest
import zope.interface


class MockObject:
    """Mock object"""


class TestAdapters(unittest.TestCase):
    def setUp(self):
        import zope.component

        context = xmlconfig.file("meta.zcml", zope.component)
        xmlconfig.file("configure.zcml", zope.component, context=context)

        import plone.event

        xmlconfig.file("configure.zcml", plone.event, context=context)

    def test_event_accessor(self):
        obj = MockObject()
        tz = pytz.timezone("Europe/Vienna")
        obj.start = datetime(2012, 12, 12, 10, 0, tzinfo=tz)
        obj.end = datetime(2012, 12, 12, 12, 0, tzinfo=tz)
        zope.interface.alsoProvides(obj, IEvent)

        # Create accessor
        acc = IEventAccessor(obj)

        # Accessor getters
        self.assertEqual(acc.start, obj.start)
        self.assertEqual(acc.end, obj.end)
        self.assertEqual(acc.duration, obj.end - obj.start)

        # Accessor setters
        start = datetime(2013, 4, 5, 16, 31, tzinfo=tz)
        end = datetime(2013, 4, 5, 16, 35, tzinfo=tz)
        acc.start = start
        acc.end = end
        self.assertTrue(acc.start == obj.start == start)
        self.assertTrue(acc.end == obj.end == end)

        # Accessor deletor
        acc.something = True
        self.assertTrue(acc.something == obj.something is True)
        del acc.something
        self.assertTrue(hasattr(acc, "something") is False)
        self.assertTrue(hasattr(obj, "something") is False)

        del acc.start
        self.assertTrue(hasattr(acc, "start") is False)
        self.assertTrue(hasattr(obj, "start") is False)
